/*
 * dblink_tds.c
 *
 * Functions returning results from a remote MSSQL database
 *
 * Diogo Biazus <diogob@gmail.com>
 * Based on the original dblink code by Joe Conway
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 *
 * IN NO EVENT SHALL THE AUTHOR OR DISTRIBUTORS BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHOR OR DISTRIBUTORS HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHOR AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHOR AND DISTRIBUTORS HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 */
#include "postgres.h"
#include "executor/spi.h"

#include "libpq-fe.h"
#include "fmgr.h"
#include "funcapi.h"
#include "executor/spi.h"
#include "utils/lsyscache.h"
#include "dblink_tds.h"

#include "tds.h"
#include "tdsconvert.h"

/*
 * Internal declarations
 */
void TDSclear(TDSSOCKET *tds);
TDSSOCKET *TDSconnect(char *server, char *user, char *password, int port, char *file_loc);
char *type_name(int type);
int exec(char *sql);

/* general utility */
#define GET_TEXT(cstrp) DatumGetTextP(DirectFunctionCall1(textin, CStringGetDatum(cstrp)))
#define GET_STR(textp) DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(textp)))

PG_FUNCTION_INFO_V1(dblink_tds);
Datum
dblink_tds(PG_FUNCTION_ARGS)
{
	FuncCallContext	*funcctx;
	TupleDesc	tupdesc = NULL;
	int		call_cntr;
	int		max_calls;
	AttInMetadata 	*attinmeta;
	bool		is_sql_cmd = false;
	char	   	*sql_cmd_status = NULL;
	MemoryContext	oldcontext;

	TDSSOCKET 	*tds;
	int 		rc;
	TDSCOLUMN 	*col;
	int 		ctype;
	CONV_RESULT 	dres;
	unsigned char 	*src;
	TDS_INT 	srclen;
	TDS_INT 	rowtype;
	TDS_INT 	resulttype;
	TDS_INT 	computeid;
	int		done_flags;
	
	/* stuff done only on the first call of the function */
	if (SRF_IS_FIRSTCALL())
	{
		Oid	functypeid;
		char	functyptype;
		Oid	funcid = fcinfo->flinfo->fn_oid;
		char	*sql = GET_STR(PG_GETARG_TEXT_P(0));
		char	*server = GET_STR(PG_GETARG_TEXT_P(1));
		char	*user = GET_STR(PG_GETARG_TEXT_P(2));
		char	*password = GET_STR(PG_GETARG_TEXT_P(3));
		int	port = 1433;
		char 	*file_loc = "/etc/freetds/freetds.conf";

		ReturnSetInfo *rsinfo = (ReturnSetInfo *) fcinfo->resultinfo;

		if (PG_NARGS() == 5)
		{
			port = PG_GETARG_INT32(4);
		}
		else if (PG_NARGS() == 6)
		{
			port = PG_GETARG_INT32(4);
			file_loc = GET_STR(PG_GETARG_TEXT_P(5));
		}

		tds = TDSconnect(server, user, password, port, file_loc);

		/* create a function context for cross-call persistence */
		funcctx = SRF_FIRSTCALL_INIT();

		/*
		 * switch to memory context appropriate for multiple function
		 * calls
		 */
		oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
		
		// Send query to MSSQL
		rc = tds_submit_query(tds, sql);
		if (rc != TDS_SUCCEED) {
			if (tds)
				TDSclear(tds);
			elog(ERROR, "tds_submit_query() failed\n");
			SRF_RETURN_DONE(funcctx);
		}
		else
		{
			if((rc = tds_process_tokens(tds, &resulttype, &done_flags, TDS_TOKEN_RESULTS)) != TDS_SUCCEED)
			{
				if (tds)
					TDSclear(tds);
				elog(ERROR, "tds_process_tokens() failed\n");
				SRF_RETURN_DONE(funcctx);
			}
		}

		// Test if is SELECT return or status return
		if(resulttype == TDS_DONE_RESULT)
		{
			is_sql_cmd = true;
			
			/* need a tuple descriptor representing one TEXT column */
			tupdesc = CreateTemplateTupleDesc(1, false);
			TupleDescInitEntry(tupdesc, (AttrNumber) 1, "status",
							   TEXTOID, -1, 0);

			/*
			 * and save a copy of the command status string to return as
			 * our result tuple
			 */
			//sql_cmd_status = tds->ret_status;
			funcctx->max_calls = 1;
		}
		else
		{
			// discard header info and get row results
			rc = tds_process_tokens(tds, &resulttype, &done_flags, TDS_TOKEN_RESULTS);
			if((rc = tds_process_tokens(tds, &rowtype, &computeid, TDS_STOPAT_ROWFMT|TDS_RETURN_DONE|TDS_RETURN_ROW|TDS_RETURN_COMPUTE)) == TDS_SUCCEED)
			{
				funcctx->max_calls = 1;
			}
		}

		/* got results, keep track of them */
		funcctx->user_fctx = tds;

		/* fast track when no results */
		if (funcctx->max_calls < 1)
		{
			if (tds)
				TDSclear(tds);
			SRF_RETURN_DONE(funcctx);
		}

		/* check typtype to see if we have a predetermined return type */
		functypeid = get_func_rettype(funcid);
		functyptype = get_typtype(functypeid);

		if (!is_sql_cmd)
		{
			if (functyptype == 'c')
				tupdesc = TypeGetTupleDesc(functypeid, NIL);
			else if (functypeid == RECORDOID)
			{
				if (!rsinfo || !IsA(rsinfo, ReturnSetInfo) ||
					rsinfo->expectedDesc == NULL)
					ereport(ERROR,
							(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
					errmsg("function returning record called in context "
						   "that cannot accept type record")));

				/* get the requested return tuple description */
				tupdesc = CreateTupleDescCopy(rsinfo->expectedDesc);
			}
			else
				/* shouldn't happen */
				elog(ERROR, "return type must be a row type");
		}

		/* store needed metadata for subsequent calls */
		attinmeta = TupleDescGetAttInMetadata(tupdesc);
		funcctx->attinmeta = attinmeta;

		MemoryContextSwitchTo(oldcontext);
	}

	/* stuff done on every call of the function */
	funcctx = SRF_PERCALL_SETUP();

	/*
	 * initialize per-call variables
	 */
	call_cntr = funcctx->call_cntr;
	max_calls = funcctx->max_calls;

	tds = (TDSSOCKET *) funcctx->user_fctx;
	attinmeta = funcctx->attinmeta;
	tupdesc = attinmeta->tupdesc;

	if (call_cntr < max_calls)	/* do when there is more left to send */
	{
		char	  **values;
		HeapTuple	tuple;
		Datum		result;
		if (!is_sql_cmd)
		{
			int			i;
			int			nfields = tds->res_info->num_cols;

			// get the field values
			values = (char **) palloc(nfields * sizeof(char *));
			for (i = 0; i < nfields; i++)
			{
				col = tds->res_info->columns[i];
				if (col->column_cur_size < 0)
				{
					values[i] = NULL;
				}
				else
				{
					ctype = tds_get_conversion_type(col->column_type, col->column_size);
					src = &(tds->res_info->current_row[col->column_offset]);
					if (is_blob_type(col->column_type))
						src = (unsigned char*) ((TDSBLOB *) src)->textvalue;
					srclen = col->column_cur_size;
					if (tds_convert(tds->tds_ctx, ctype, (TDS_CHAR*) src, srclen, SYBVARCHAR, &dres) < 0)
						continue;

					values[i] = (char *) dres.c;
				}
			}
		}
		else
		{
			values = (char **) palloc(1 * sizeof(char *));
			values[0] = sql_cmd_status;
		}

		/* build the tuple */
		tuple = BuildTupleFromCStrings(attinmeta, values);

		/* make the tuple into a datum */
		result = HeapTupleGetDatum(tuple);

		// If there are more rows keep executing
		if((rc = tds_process_tokens(tds, &rowtype, &computeid, TDS_STOPAT_ROWFMT|TDS_RETURN_ROW|TDS_RETURN_COMPUTE)) == TDS_SUCCEED)
		{
			funcctx->call_cntr--;
		}

		SRF_RETURN_NEXT(funcctx, result);
	}
	else
	{
		/* do when there is no more left */
		TDSclear(tds);
		SRF_RETURN_DONE(funcctx);
	}
}

PG_FUNCTION_INFO_V1(dblink_tds_create_view);
Datum
dblink_tds_create_view(PG_FUNCTION_ARGS)
{
	TDSSOCKET *tds;
	int rc;
	TDSCOLUMN *col;
	int nfields;
	TDS_INT resulttype;

	char	*table = GET_STR(PG_GETARG_TEXT_P(0));
	char	*view_name = GET_STR(PG_GETARG_TEXT_P(1));
	char	*server = GET_STR(PG_GETARG_TEXT_P(2));
	char	*user = GET_STR(PG_GETARG_TEXT_P(3));
	char	*password = GET_STR(PG_GETARG_TEXT_P(4));

	int	port = 1433;
	char 	*file_loc = "/etc/freetds/freetds.conf";
	char 	*format = "SELECT * FROM %s";
	char	*sql;
	int 	i;
	char	*create_view;
	char	*create_view_format = "CREATE VIEW %s AS %s";
	char	*create_view_sql;
	char	*create_view_sql_format;
	char	*create_view_inner_sql;
	
	int	fields_string_size = 0;
	int	types_string_size = 0;
	int	done_flags;

	sql = palloc(strlen(table) + 14);
	sprintf(sql, format, table);

	if (PG_NARGS() == 6)
	{
		port = PG_GETARG_INT32(5);
	}
	else if (PG_NARGS() == 7)
	{
		port = PG_GETARG_INT32(5);
		file_loc = GET_STR(PG_GETARG_TEXT_P(6));
	}

	tds = TDSconnect(server, user, password, port, file_loc);

	// Send query to MSSQL
	rc = tds_submit_query(tds, sql);
	if (rc != TDS_SUCCEED) {
		if (tds)
			TDSclear(tds);
		elog(ERROR, "tds_submit_query() failed\n");
	}
	else
	{
		if((rc = tds_process_tokens(tds, &resulttype, &done_flags, TDS_TOKEN_RESULTS)) != TDS_SUCCEED)
		{
			if (tds)
				TDSclear(tds);
			elog(ERROR, "tds_process_tokens() failed\n");
		}
	}

	if(!(rc = tds_process_tokens(tds, &resulttype, &done_flags, TDS_TOKEN_RESULTS)))
		elog(ERROR, "tds_process_tokens() failed\n");	

	// Process fields to get size of fields and types strings
	nfields = tds->res_info->num_cols;
	for (i = 0; i < nfields; i++)
	{
		col = tds->res_info->columns[i];
		fields_string_size += strlen(col->column_name) + 2;
		types_string_size += strlen(type_name(col->column_type));
	}

	// Allocate and start to concatenate the format to the create view command
	create_view_sql_format = palloc(fields_string_size + types_string_size + 62);
	create_view_sql_format = strcat(create_view_sql_format, "SELECT * FROM dblink_tds('%s', '%s', '%s', '%s', %i, '%s') a(");

	// Allocate and start to concatenate inner sql command to be sent to mssql
	create_view_inner_sql = palloc(fields_string_size + strlen(table) + 12);
	create_view_inner_sql = strcat(create_view_inner_sql, "SELECT ");

	for (i = 0; i < nfields; i++)
	{
		col = tds->res_info->columns[i];
		create_view_sql_format = strcat(create_view_sql_format, col->column_name);
		create_view_sql_format = strcat(create_view_sql_format, " ");
		create_view_sql_format = strcat(create_view_sql_format, type_name(col->column_type));

		create_view_inner_sql = strcat(create_view_inner_sql, col->column_name);

		// If is not in last field append a comma
		if(i != (nfields -1))
		{
			create_view_sql_format = strcat(create_view_sql_format, ",");
			create_view_inner_sql = strcat(create_view_inner_sql, ", ");
		}
	}
	// Finish the concatenations
	create_view_sql_format = strcat(create_view_sql_format, ")");
	create_view_inner_sql = strcat(create_view_inner_sql, " FROM ");
	create_view_inner_sql = strcat(create_view_inner_sql, table);

	// Creates the sql used in create view command
	create_view_sql = palloc(strlen(create_view_sql_format) + strlen(create_view_inner_sql) + strlen(server) + strlen(user) + strlen(password) + 6 + strlen(file_loc));
	sprintf(create_view_sql, create_view_sql_format, create_view_inner_sql, server, user, password, port, file_loc);

	// Creates the create view command
	create_view = palloc(strlen(view_name) + strlen(create_view_sql) + 16);
	sprintf(create_view, create_view_format, view_name, create_view_sql);
	
	// Free memory
	pfree(create_view_sql_format);
	pfree(create_view_inner_sql);
	pfree(create_view_sql);

	if(exec(create_view) > -1)
		PG_RETURN_TEXT_P(GET_TEXT("OK"));
	else
		PG_RETURN_TEXT_P(GET_TEXT("ERROR"));
}

PG_FUNCTION_INFO_V1(dblink_tds_create_rules);
Datum
dblink_tds_create_rules(PG_FUNCTION_ARGS)
{
	char	*table = GET_STR(PG_GETARG_TEXT_P(0));
	char	*view_name = GET_STR(PG_GETARG_TEXT_P(1));
	char	*server = GET_STR(PG_GETARG_TEXT_P(2));
	char	*user = GET_STR(PG_GETARG_TEXT_P(3));
	char	*password = GET_STR(PG_GETARG_TEXT_P(4));

	int	port = 1433;
	char 	*file_loc = "/etc/freetds/freetds.conf";

	int	ret;
	int 	proc;
	char 	*sql;
	int 	i;

	if (PG_NARGS() == 6)
	{
		port = PG_GETARG_INT32(5);
	}
	else if (PG_NARGS() == 7)
	{
		port = PG_GETARG_INT32(5);
		file_loc = GET_STR(PG_GETARG_TEXT_P(6));
	}
	
	sql = palloc(strlen(view_name) + 14);
	sprintf(sql, "SELECT * FROM %s", view_name);

	SPI_connect();
	ret = SPI_exec(sql, 1);
	proc = SPI_processed;

	if (ret == SPI_OK_SELECT && SPI_processed > 0)
	{
		TupleDesc tupdesc = SPI_tuptable->tupdesc;
		SPITupleTable *tuptable = SPI_tuptable;	
		HeapTuple tuple = tuptable->vals[0];
		for (i = 1; i <= tupdesc->natts; i++)
		{
			
		}
        }
	SPI_finish();

	PG_RETURN_TEXT_P(GET_TEXT("OK"));
}

/*************************************************************
 * internal functions
 */

/*
 * exec
 *
 * Executes the given sql query (does not work with selects)
 */
int
exec(char *sql)
{
    int ret;
    SPI_connect();
    ret = SPI_exec(sql, 0);
    SPI_finish();
    return ret;
}

/*
 * type_name
 *
 * Return the PostgreSQL type name for a given MSSQL server type number
 */
char
*type_name(int type)
{
	switch(type)
	{
		case 38:
			return "bigint";
			break;
		case 39:
			return "varchar";
			break;
		case 47:
			return "char";
			break;
		case 50:
			return "boolean";
			break;
		case 56:
			return "integer";
			break;
		case 106:
		case 108:
			return "numeric";
			break;
		case 109:
			return "float";
			break;
		case 110:
			return "money";
			break;
		case 111:
			return "timestamp";
			break;
		default:
			return "text";
			break;
	}
}

/*
 * TDSclear
 *
 * Free all structures used in FreeTDS connection
 */
void
TDSclear(TDSSOCKET *tds)
{
	/* close up shop */
	tds_free_socket(tds);
	tds_free_context((TDSCONTEXT *) tds->tds_ctx);
}

/*
 * TDSconnect
 *
 * Opens FreeTDS connection and return a TDSSOCKET
 */

TDSSOCKET
*TDSconnect(char *server, char *user, char *password, int port, char *file_loc)
{
	TDSLOGIN *login;
	TDSCONTEXT *context;
	TDSCONNECTION *connection;
	TDSSOCKET *tds;

	/* grab a login structure */
	login = tds_alloc_login();

	tds_set_user(login, user);
	tds_set_server(login, server);
	tds_set_port(login, port);
	tds_set_passwd(login, password);
	tds_set_interfaces_file_loc(file_loc);

	// Set constant parameters
	tds_set_app(login, DBLINK_TDS_APP);
	tds_set_library(login, DBLINK_TDS_LIBRARY);
	tds_set_packet(login, DBLINK_TDS_PACKET);
	
	context = tds_alloc_context(NULL);

	if (context->locale && !context->locale->date_fmt) {
		context->locale->date_fmt = strdup("%b %e %Y %I:%M%p");
	}

	tds = tds_alloc_socket(context, 512);
	tds_set_parent(tds, NULL);
	connection = tds_read_config_info(NULL, login, context->locale);
	if (!connection || tds_connect(tds, connection) == TDS_FAIL) {
		elog(ERROR, "There was a problem connecting to the server\n");
	}

	tds_free_login(login);
	return tds;
}
